from datetime import datetime, timedelta
from typing import Dict, List

import pandas as pd
from dateutil.parser import parse


def format_tank(readings: List[Dict], freq: str = "30min") -> pd.DataFrame or None:

    if not readings:
        return None

    validated = []

    df = pd.DataFrame(readings)
    df = (
        df[["read_time", "volume"]]
        .set_index("read_time")
        .resample(timedelta(minutes=30))
        .agg(["mean"])
        .interpolate(limit=4)
    )
    df.reset_index(inplace=True)
    df.columns = df.columns.get_level_values(0)
    readings = df.to_dict("records")

    for i in range(len(readings)):
        if "read_time" in readings[i] and "volume" in readings[i]:
            validated.append(readings[i])

    if not validated:
        return None

    validated = sorted(validated, key=lambda x: x["read_time"])
    validated = [d for d in validated if d["volume"]]

    df = pd.DataFrame(validated)
    df.read_time = pd.to_datetime(df.read_time)
    df["y"] = df.volume.diff() * -1
    df = df[["read_time", "y"]].rename(columns={"read_time": "ds"})
    df = df[df.y < 800].copy()
    df.y = df.y.clip(lower=0)
    df = df.set_index("ds").groupby(pd.Grouper(freq=freq)).sum()

    if len(df) == 2:
        return None

    df.reset_index(level=0, inplace=True)
    output = df.iloc[1:-1]

    return output


def format_sales(sales_input: List[Dict]) -> pd.DataFrame or None:

    if sales_input == [] or sales_input is None:
        return None

    sales_input = sorted(sales_input, key=lambda x: x["date"])
    sales_df = pd.DataFrame(sales_input)
    df = sales_df[["date", "sales"]]

    df2 = df.copy()
    df2["sales"] = df["sales"].clip(
        lower=df.sales.quantile(0.015), upper=df.sales.quantile(0.995)
    )

    df3 = df2.rename(columns={"date": "ds", "sales": "y"})

    result = df3.copy()
    result["ds"] = df3.loc[:, "ds"].astype("datetime64[ns]")

    return result


def gen_past(df: pd.DataFrame, freq: str = "30min", periods: int = 432) -> pd.DataFrame:

    if len(df) == 0 or df is None:
        now = datetime.now()
        end_0 = now - (now - datetime.min) % timedelta(minutes=30)

    else:
        end_0 = df.iloc[-1, 0]

    past = pd.date_range(end=end_0, freq=freq, periods=periods)
    past = pd.DataFrame(past)
    past.rename(columns={0: "ds"}, inplace=True)
    past["ds"] = pd.to_datetime(past["ds"], format="%Y-%m-%d")

    return past


if __name__ == "__main__":
    import json

    with open("../fisher.json") as f:
        data = json.load(f)
    for row in data:
        row["read_time"] = parse(row["read_time"])
    f = format_tank(data)
    print(f)



from datetime import datetime, timedelta
from typing import List

import numpy as np
import pandas as pd
from prophet import Prophet
from sparkles import trunc_date
from sparkles.fbprophet_hacks import suppress_stdout_stderr


def forecast_near(
    tank_history: pd.DataFrame = None,
    start_time=None,
    forecast_freq: str = None,
    forecast_length: int = None,
):

    if tank_history is None or len(tank_history) < 432:
        return

    if start_time is None:
        now = datetime.now()
        now_hh = now - (now - datetime.min) % timedelta(minutes=30)
        start_time = now_hh

    if forecast_freq is None:
        forecast_freq = "30min"

    if forecast_length is None:
        forecast_length = 144

    future = pd.date_range(
        start=start_time, freq=forecast_freq, periods=forecast_length
    )
    future = pd.DataFrame(future)
    future.rename(columns={0: "ds"}, inplace=True)

    m = Prophet(
        changepoint_prior_scale=0.05,
        daily_seasonality=True,
        weekly_seasonality=True,
        yearly_seasonality=False,
    )
    with suppress_stdout_stderr():
        m.fit(tank_history)
    forecast = m.predict(future)

    output = forecast.loc[:, ["ds", "yhat_lower", "yhat_upper", "yhat"]]
    output.rename(columns={"yhat_lower": "lower", "yhat_upper": "upper"}, inplace=True)

    for field in ["yhat", "lower", "upper"]:
        output.loc[output[field] < 0, field] = 0

    return output


def forecast_near_approx(
    daily_lift_est: int,
    hh_seasonality: List,
    start_time=None,
    forecast_freq: str = None,
    forecast_length: int = None,
):
    if daily_lift_est is None:
        return None

    if hh_seasonality is None:
        return None

    if start_time is None:
        now = datetime.now()
        now_hh = now - (now - datetime.min) % timedelta(minutes=30)
        start_time = now_hh

    if forecast_freq is None:
        forecast_freq = "30min"

    if forecast_length is None:
        forecast_length = 144

    predictions = pd.DataFrame(columns=["ds", "ts", "base", "estimate"])
    predictions["ds"] = pd.date_range(
        pd.to_datetime(start_time), freq=forecast_freq, periods=forecast_length
    )

    predictions["ts"] = predictions.ds.dt.time
    predictions["base"] = daily_lift_est * (1 / 48)

    hh_multiplier = pd.DataFrame(columns=["ts", "value"])
    hh_multiplier["ts"] = pd.date_range("00:00", "23:30", freq="30min")
    hh_multiplier["ts"] = hh_multiplier.ts.dt.time
    hh_multiplier["value"] = hh_seasonality

    merged = pd.merge(left=predictions, right=hh_multiplier)

    merged["estimate"] = merged["base"] * merged["value"]

    output = merged[["ds", "estimate"]].copy()
    output.rename(columns={"estimate": "yhat"}, inplace=True)

    output["lower"] = output["yhat"] - 2 * output["yhat"].std()
    output["upper"] = output["yhat"] + 2 * output["yhat"].std()

    for field in ["yhat", "lower", "upper"]:
        output.loc[output[field] < 0, field] = 0

    result = output.sort_values(by="ds")

    return result


def forecast_far(
    daily_lift_est: int,
    weekly_seasonality: List,
    yearly_seasonality: List = None,
    start_date=None,
    forecast_length: int = None,
):
    if daily_lift_est is None:
        return None

    if yearly_seasonality is None:
        yearly_seasonality = [1] * 53

    if weekly_seasonality is None:
        return None

    if start_date is None:
        start_date = trunc_date(datetime.now())

    if forecast_length is None:
        forecast_length = 90

    yearly_seasonality = pd.Series(yearly_seasonality)
    weekly_seasonality = pd.Series(weekly_seasonality)

    predictions = pd.DataFrame(columns=["woy", "dow", "base", "weekly", "daily"])
    predictions["woy"] = np.repeat(range(1, 54), 7)  # 52.14 'week of year' -> 53
    predictions["dow"] = [0, 1, 2, 3, 4, 5, 6] * 53  # 7 'day of week' each week of year
    predictions["base"] = daily_lift_est

    weekly = np.repeat(predictions.base.iloc[1] * 7 * yearly_seasonality, 7)
    weekly.index = range(0, 371)
    predictions["weekly"] = weekly

    weekly_seas_repeated = pd.concat([weekly_seasonality] * 53, ignore_index=True)
    predictions["daily"] = predictions.weekly * (1 / 7) * weekly_seas_repeated

    future = pd.date_range(start=start_date, freq="1D", periods=forecast_length)
    future = pd.DataFrame(future)
    future.rename(columns={0: "ds"}, inplace=True)

    future["dow"] = future.ds.dt.weekday  # future.ds.dt.dayofweek
    future["woy"] = future.ds.dt.isocalendar().week

    output = pd.merge(
        future, predictions, left_on=["woy", "dow"], right_on=["woy", "dow"]
    )

    output = output[["ds", "daily"]]
    output.rename(columns={"daily": "yhat"}, inplace=True)

    output["lower"] = output["yhat"] - 2 * output["yhat"].std()
    output["upper"] = output["yhat"] + 2 * output["yhat"].std()

    for field in ["yhat", "lower", "upper"]:
        output.loc[output[field] < 0, field] = 0

    return output


if __name__ == "__main__":
    pass



import numpy as np
import pandas as pd
from prophet import Prophet
from sparkles.fbprophet_hacks import suppress_stdout_stderr

# declare variables for testing
store = "KT220"
tank = "7"
fc_url = "https://fc.bbdev.host3.capspire.com"


def decompose_sales(sales_data: pd.DataFrame):

    if sales_data is None or len(sales_data) < 7:
        return pd.Series([1] * 53), pd.Series([1] * 7)

    m = Prophet(
        changepoint_prior_scale=0.05,
        daily_seasonality=False,
        weekly_seasonality=True,
        yearly_seasonality=True,
        seasonality_mode="multiplicative",
    )
    with suppress_stdout_stderr():
        m.fit(sales_data)

    future = m.make_future_dataframe(periods=1, freq="1D")
    forecast = m.predict(future)
    forecast_reduced = forecast.loc[:, ["ds", "yearly", "weekly"]]
    forecast_reduced["dow"] = forecast_reduced.ds.dt.weekday
    forecast_reduced["doy"] = forecast_reduced.ds.dt.dayofyear
    forecast_reduced["woy"] = forecast_reduced.ds.dt.isocalendar().week

    weekly_seasonality = forecast_reduced.groupby("dow")["weekly"].mean() + 1

    if len(weekly_seasonality) < 7:
        return pd.Series([1] * 53), pd.Series([1] * 7)

    if len(sales_data) < 350:
        return pd.Series([1] * 53), weekly_seasonality

    forecast_reduced = forecast_reduced.iloc[-366:]
    forecast_reduced.sort_values(by=["woy", "dow"], inplace=True)
    year_trend = forecast_reduced.groupby("woy").mean()
    year_trend = year_trend["yearly"] + 1

    yearly_seasonality = ensure_53_weeks(year_trend)

    if len(weekly_seasonality) == 7 and len(yearly_seasonality) == 53:
        return yearly_seasonality, weekly_seasonality

    return pd.Series([1] * 53), pd.Series([1] * 7)


def ensure_53_weeks(initial: pd.DataFrame) -> pd.DataFrame:
    """
    takes a shorter dataframe and ensures at least 53 rows by wrapping the first rows onto the end
    """
    extra_weeks_needed = 53 - len(initial)
    if extra_weeks_needed > 0:
        ret = pd.concat([initial, initial[:extra_weeks_needed]], ignore_index=True)
    else:
        ret = initial.copy()
    return ret


def decompose_seasonality(tank_history: pd.DataFrame):
    if tank_history is None or len(tank_history) < 48:
        return main_hh_seas, main_dow_seas

    m = Prophet(
        changepoint_prior_scale=0.05,
        daily_seasonality=True,
        weekly_seasonality=True,
        yearly_seasonality=False,
        seasonality_mode="multiplicative",
    )

    with suppress_stdout_stderr():
        m.fit(tank_history)

    future = m.make_future_dataframe(periods=1, freq="30min")
    forecast = m.predict(future)

    forecast_reduced = forecast.loc[:, ["ds", "daily", "weekly"]]
    forecast_reduced["time"] = forecast_reduced.ds.dt.time
    forecast_reduced["dow"] = forecast_reduced.ds.dt.weekday
    forecast_reduced.drop(columns="ds", inplace=True)

    day_seas = forecast_reduced.groupby("time").daily.mean()
    week_seas = forecast_reduced.groupby("dow").weekly.mean()

    day_seas.sort_index(inplace=True)
    week_seas.sort_index(inplace=True)

    if len(day_seas) != 48 and len(week_seas) != 7:
        return main_hh_seas, main_dow_seas

    if len(day_seas) != 48 or len(week_seas) != 7:
        if len(week_seas) != 7:
            return list(day_seas + 1), main_dow_seas
        if len(day_seas) != 48:
            return main_hh_seas, list(week_seas + 1)

    return list(day_seas + 1), list(week_seas + 1)


main_hh_seas = [
    1.17,
    0.94,
    0.74,
    0.57,
    0.45,
    0.37,
    0.32,
    0.30,
    0.28,
    0.25,
    0.21,
    0.16,
    0.11,
    0.05,
    0.02,
    0.01,
    0.03,
    0.08,
    0.17,
    0.28,
    0.41,
    0.54,
    0.68,
    0.81,
    0.92,
    1.04,
    1.15,
    1.25,
    1.35,
    1.44,
    1.51,
    1.57,
    1.62,
    1.65,
    1.67,
    1.69,
    1.72,
    1.75,
    1.80,
    1.86,
    1.93,
    1.98,
    2.00,
    1.98,
    1.91,
    1.79,
    1.61,
    1.40,
]

truck_hh_seas = [
    0.78,
    0.64,
    0.51,
    0.42,
    0.35,
    0.31,
    0.28,
    0.28,
    0.27,
    0.28,
    0.26,
    0.23,
    0.21,
    0.20,
    0.19,
    0.22,
    0.27,
    0.37,
    0.50,
    0.66,
    0.85,
    1.04,
    1.18,
    1.33,
    1.45,
    1.53,
    1.55,
    1.56,
    1.56,
    1.55,
    1.52,
    1.52,
    1.53,
    1.52,
    1.54,
    1.56,
    1.58,
    1.59,
    1.59,
    1.57,
    1.54,
    1.51,
    1.47,
    1.37,
    1.31,
    1.19,
    1.06,
    0.92,
]

main_dow_seas = [1.04, 0.98, 0.98, 1, 1.08, 1.01, 0.9]
truck_dow_seas = [1.11, 1.12, 1.09, 1.06, 1.03, 0.75, 0.72]


def get_tank_readings():
    import requests

    headers = {
        "accept": "application/json",
    }

    params = (
        ("start_date", "2020-10-01 00:00:00"),
        ("end_date", "2020-10-18 00:00:00"),
        ("store_number", "103"),
        ("tank_id", "1"),
        ("system_psk", "7636d15a00204b62c56da0d2311ce69ae52ef28cc5940545"),
    )

    response = requests.post(
        "https://ktdev.bb.gravitate.energy/service/ims/tank_inventory/readings",
        headers=headers,
        params=params,
    )
    return response.json()


if __name__ == "__main__":
    from tank_forecaster.validation import format_tank

    tr = get_tank_readings()
    val = format_tank(tr)
    day_seas, week_seas = decompose_seasonality(val)
    print(day_seas)
    print(week_seas)



from tank_forecaster import decomp


def seas_are_of_correct_length(x):
    return len(x[0]) == 53 and len(x[1]) == 7


def test_decompose_sales_none_returns_none():
    x = decomp.decompose_sales(None)
    assert seas_are_of_correct_length(x)


def test_decompose_sales_little_data_returns_lists(sales_little_data):
    x = decomp.decompose_sales(sales_little_data)
    assert seas_are_of_correct_length(x)


def test_decompose_sales_week_data_returns_correct_len(sales_week_data):
    x = decomp.decompose_sales(sales_week_data)
    assert seas_are_of_correct_length(x)


def test_decompose_sales_three_week_data_returns_correct_len(sales_three_week_data):
    x = decomp.decompose_sales(sales_three_week_data)
    assert seas_are_of_correct_length(x)
